import logging
from abc import abstractmethod, ABC
from typing import Optional, Generic

from typing_extensions import TypeVar

from diskurs.entities import ChatMessage, PromptArgument, MessageType, Role
from diskurs.logger_setup import get_logger
from diskurs.protocols import ConversationDispatcher, LLMClient, Agent, ConversationParticipant, Conversation, Prompt

logger = logging.getLogger(__name__)

Prompt = TypeVar("Prompt", bound=Prompt)

# TODO: implement conditional rendering i.e. for each agent, only the information relevant to it is shown


class BaseAgent(ABC, Agent, ConversationParticipant, Generic[Prompt]):
    def __init__(
        self,
        name: str,
        prompt: Prompt,
        llm_client: LLMClient,
        topics: Optional[list[str]] = None,
        dispatcher: Optional[ConversationDispatcher] = None,
        max_trials: int = 5,
    ):
        self.dispatcher = dispatcher
        self.prompt = prompt
        self.name = name
        self._topics = topics or []
        self.max_trials = max_trials
        self.llm_client = llm_client
        self.logger = get_logger(f"diskurs.agent.{self.name}")

        self.logger.info(f"Initializing agent {self.name}")

    @property
    def topics(self) -> list[str]:
        return self._topics

    @topics.setter
    def topics(self, value):
        self._topics = value

    @abstractmethod
    def invoke(self, conversation: Conversation | str) -> Conversation:
        """
        Runs the agent on a conversation, performing reasoning steps until the user prompt is final,
        meaning all the conditions, as specified in the prompt's is_final function, are met.
        If the conversation is a string i.e. starting a new conversation, the agent will prepare
        the conversation by setting the user prompt argument's content to this string.

        :param conversation: The conversation object to run the agent on. If a string is provided, the agent will
            start a new conversation with the string as the user query's content.
        :return: the updated conversation object after the agent has finished reasoning. Contains
            the chat history, with all the system and user messages, as well as the final answer.
        """
        pass

    def register_dispatcher(self, dispatcher: ConversationDispatcher) -> None:
        self.dispatcher = dispatcher

        self.logger.debug(f"Registered dispatcher {dispatcher} for agent {self.name}")

    @abstractmethod
    def process_conversation(self, conversation: Conversation) -> None:
        """
        Receives a conversation from the dispatcher, i.e. message bus, processes it and finally publishes
        a deep copy of the resulting conversation back to the dispatcher.

        :param conversation: The conversation object to process.
        """
        pass

    def prepare_conversation(
        self,
        conversation: Conversation,
        system_prompt_argument: PromptArgument,
        user_prompt_argument: PromptArgument,
        message_type: MessageType = MessageType.CONVERSATION,
    ) -> Conversation:
        """
        Ensures the conversation is in a valid state by creating a new set of prompts
        and prompt_variables for system and user, as well creating a fresh copy of the conversation.

        :param conversation: A conversation object, possible passed from another agent
            or a string to start a new conversation.
        :param system_prompt_argument: The system prompt argument to use for the system prompt.
        :param user_prompt_argument: The user prompt argument to use for the user prompt.
        :param message_type: The type of message to render the user prompt as.
        :return: A deep copy of the conversation, in a valid state for this agent
        """
        self.logger.debug(f"Preparing conversation for agent {self.name}")
        system_prompt = self.prompt.render_system_template(self.name, prompt_args=system_prompt_argument)
        user_prompt = self.prompt.render_user_template(
            name=self.name, prompt_args=user_prompt_argument, message_type=message_type
        )

        return conversation.update(
            system_prompt_argument=system_prompt_argument,
            user_prompt_argument=user_prompt_argument,
            system_prompt=system_prompt,
            user_prompt=user_prompt,
            active_agent=self.name,
        )

    def return_fail_validation_message(self, response):
        return response.append(
            ChatMessage(
                role=Role.USER, content=f"No valid answer found after {self.max_trials} trials.", name=self.name
            )
        )

    def generate_validated_response(
        self, conversation: Conversation, message_type: MessageType = MessageType.CONVERSATION
    ) -> Conversation:
        response = None

        for max_trials in range(self.max_trials):
            self.logger.debug(f"Generating validated response trial {max_trials + 1} for Agent {self.name}")

            response = self.llm_client.generate(conversation, getattr(self, "tools", None))

            parsed_response = self.prompt.parse_user_prompt(
                llm_response=response.last_message.content,
                old_user_prompt_argument=response.user_prompt_argument,
                message_type=message_type,
            )

            if isinstance(parsed_response, PromptArgument):
                self.logger.debug(f"Valid response found for Agent {self.name}")
                return response.update(
                    user_prompt_argument=parsed_response,
                    user_prompt=self.prompt.render_user_template(name=self.name, prompt_args=parsed_response),
                )
            elif isinstance(parsed_response, ChatMessage):
                self.logger.debug(f"Invalid response, created corrective message for Agent {self.name}")

                conversation = response.update(user_prompt=parsed_response)
            else:
                self.logger.error(f"Failed to parse response from LLM model: {parsed_response}")
                raise ValueError(f"Failed to parse response from LLM model: {parsed_response}")

        return self.return_fail_validation_message(response or conversation)
